package com.capitalware.logging;

import java.io.File;
import java.io.FileDescriptor;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.nio.file.Files;
import java.nio.file.LinkOption;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.attribute.BasicFileAttributeView;
import java.nio.file.attribute.BasicFileAttributes;
import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;

/**
 * This class will handle logging information to a file.
 * 
 * <ul>Supported JVM Environment variables:
 * <li><B><i>SimpleLogger.LogFileName</i></B> - any valid file name</li>
 * <li><B><i>SimpleLogger.LogDirectory</i></B> - any valid local directory</li>
 * <li><B><i>SimpleLogger.IncludePackageName</i></B> - valid values are: true or false</li>
 * <li><B><i>SimpleLogger.LogLevel</i></B> - valid values are: debug, info, warn or error</li>
 * <li><B><i>SimpleLogger.MaxBackupFiles</i></B> - any valid integer</li>
 * <li><B><i>SimpleLogger.MaxFileSize</i></B> - any valid integer</li>
 * <li><B><i>SimpleLogger.LogRotationType</i></B> - valid values are: daily or size</li>
 *  </ul>
 *
 * @author Roger Lacroix, Capitalware Inc.
 * @version 1.1.0.0
 * @license Apache 2 License
 */
public class SimpleLogger
{
   private static final DateTimeFormatter LOGGER_TIMESTAMP = DateTimeFormatter.ofPattern("yyyy/MM/dd HH:mm:ss.SSS");
   private static final DateTimeFormatter JULIAN_DATE      = DateTimeFormatter.ofPattern("yyyyDDD");
   private static final String            HEX              = "0123456789ABCDEF";

   /**
    * JVM environment variables names that can be used.
    */
   private static final String            ENV_LOG_FILENAME        = "SimpleLogger.LogFileName";
   private static final String            ENV_LOG_DIRECTORY       = "SimpleLogger.LogDirectory";
   private static final String            ENV_INCLUDE_PACKAGE_NAME= "SimpleLogger.IncludePackageName";
   private static final String            ENV_LOG_LEVEL           = "SimpleLogger.LogLevel";
   private static final String            ENV_MAX_BACKUP_FILES    = "SimpleLogger.MaxBackupFiles";
   private static final String            ENV_MAX_FILE_SIZE       = "SimpleLogger.MaxFileSize";
   private static final String            ENV_LOG_ROTATION_TYPE   = "SimpleLogger.LogRotationType";

   private static FileOutputStream  fos = null;
   private static PrintStream       ps = null;
   
   private static String            logFileName = "logger.log";
   private static String            pathAndFileName = logFileName;
   private static LogLevel          level = LogLevel.INFO;
   private static LogRotationType   rotationType = LogRotationType.DAILY;
   private static int               maxBackupFiles = 9;
   private static int               maxFileSize = 10 * 1024 * 1024; // 10MB default value
   private static boolean           includePackageName = true;
   private static int               julianDate = 0;

   /**
    * Supported log levels. 
    * lowest to highest log levels: ERROR < WARN < INFO < DEBUG
    */
   public static enum LogLevel 
   {
      /**
       * Included all output from error, warn, info & debug.
       */
      DEBUG,
      /**
       * Included all output from error, warn & info.
       */
      INFO,
      /**
       * Included all output from error & warn.
       */
      WARN,
      /**
       * Only included output from error.
       */
      ERROR
   }

   /**
    * Supported types of log file rotations.
    */
   public static enum LogRotationType 
   {
      /**
       * Daily log rotation.  The first entry after midnight causes the log rotation.
       */
      DAILY, 
      /**
       * Max size log rotation.  If the existing log file size is larger than the max file size then the log rotation is done.
       */
      SIZE   
   }

   /**
    * Create the directory, if necessary, construct the logPathAndFileName and the call 
    * handleSettingPrintStream method to open/create the logfile.
    * The log file name must be set as the JVM environment variable ("SimpleLogger.LogFileName") or else it will throw an exception. 
    *
    * @throws IllegalArgumentException
    */
   public static void setLogFile() throws IllegalArgumentException
   {
      setLogFile(System.getProperty(ENV_LOG_FILENAME), System.getProperty(ENV_LOG_DIRECTORY));
   }

   /**
    * Create the directory, if necessary, construct the logPathAndFileName and the call 
    * handleSettingPrintStream method to open/create the logfile.
    *
    * @param logFileName The name of the log file
    * @throws IllegalArgumentException
    */
   public static void setLogFile(String logFileName) throws IllegalArgumentException
   {
      setLogFile(logFileName, System.getProperty(ENV_LOG_DIRECTORY));
   }
   
   /**
    * Create the directory, if necessary, construct the logPathAndFileName and the call 
    * handleSettingPrintStream method to open/create the logfile.
    *
    * @param logFileName The name of the log file
    * @param logDirectory The directory where to write the log file to. Default is current working directory.
    * @throws IllegalArgumentException
    */
   public static void setLogFile(String logFileName, String logDirectory) throws IllegalArgumentException
   {
      SimpleLogger.logFileName = logFileName;
      julianDate = LocalDate.now().getYear() * 1000 + LocalDate.now().getDayOfYear(); 

      retrieveEnvironmentVariables();
      
      try
      {
         // Make sure the directory tree has been created.
         if ( (logDirectory!= null) && (!((new File(logDirectory)).exists())) )
            (new File(logDirectory)).mkdirs(); // create empty directory tree
      }
      catch(Exception e)
      {
         e.printStackTrace();
      }

      if (logDirectory != null)
      {    
         if (logDirectory.trim().lastIndexOf(System.getProperty("file.separator"))  == (logDirectory.trim().length() - 1))
            pathAndFileName = logDirectory.trim() + logFileName.trim();
         else
            pathAndFileName = logDirectory.trim() + System.getProperty("file.separator") + logFileName.trim();
      }
      else
      {
         if ( (logFileName != null) && !("".equals(logFileName.trim())) )
            pathAndFileName = System.getProperty("user.dir")+System.getProperty("file.separator")+logFileName.trim();
         else
            throw new IllegalArgumentException("logFileName parameter cannot be null or blank.");
      }

      /*
       * If the logfile already exists then we may need to rotate it.
       */
      if ( (new File(pathAndFileName)).exists() )
      {
         if (rotationType == LogRotationType.DAILY)
         {
            try
            {
               if (julianDate != Integer.parseInt(getLastModified((new File(pathAndFileName))).format(JULIAN_DATE)))
                  moveLogFiles();
            }
            catch (Exception e)
            {
               e.printStackTrace();
            }
         }
         else // LogRotationType.SIZE
         {
            if ((new File(pathAndFileName)).length() > maxFileSize)
               moveLogFiles();
         }
      }
      
      handleSettingPrintStream();
   }

   /**
    * Handle the opening of the log file and assigning it to the system's standard out and standard error.
    * @throws IllegalArgumentException
    */
   private static void handleSettingPrintStream() throws IllegalArgumentException
   {
      boolean fail1 = false;
      boolean fail2 = false;

      try
      {
         // open the file
         fos = new FileOutputStream(pathAndFileName, true);
         // get it a stream
         ps = new PrintStream(fos, true);
      }
      catch(Exception e)
      {
         fail1 = true;
         e.printStackTrace();

         // Try the current directory.
         if ( (logFileName != null) && !("".equals(logFileName.trim())) )
            pathAndFileName = System.getProperty("user.dir")+System.getProperty("file.separator")+logFileName.trim();
         else
            throw new IllegalArgumentException("logFileName parameter cannot be null or blank.");

         try
         {
            // open the file
            fos = new FileOutputStream(pathAndFileName, true);
            // get it a stream
            ps = new PrintStream(fos, true);
         }
         catch(Exception e1)
         {
            fail2 = true;
            e.printStackTrace();
         }
      }

      /**
       * If both attempts to open a logfile fails then just use the default logging to standard out
       */
      if ( (fail1) && (fail2) )
      {
         System.err.println("Warning: Continuing without log file");
      }

      if (ps != null)
      {
         try
         {
            /*
             * Set the opened print stream for all System.out.* and System.err.* method calls. 
             */
            System.setOut(ps);
            System.setErr(ps);
         }
         catch(Exception e)
         {
            e.printStackTrace();
         }
      }
   }

   /**
    * Retrieve any JVM environment variables that may have been set by the user.
    * 
    * SimpleLogger.LogDirectory - any valid local directory
    * SimpleLogger.LogLevel - valid values are: debug, info, warn or error
    * SimpleLogger.MaxBackupFiles - any valid integer
    * SimpleLogger.MaxFileSize - any valid integer
    * SimpleLogger.LogRotationType - valid values are: daily or size
    * SimpleLogger.IncludePackageName - valid values are: true or false
    */
   private static void retrieveEnvironmentVariables()
   {
      if (System.getProperty(ENV_LOG_LEVEL) != null)
      {
         String temp = System.getProperty(ENV_LOG_LEVEL);
         if ("debug".equals(temp.trim().toLowerCase()))
            level = LogLevel.DEBUG;
         else if ("info".equals(temp.trim().toLowerCase()))
            level = LogLevel.INFO;
         else if ("warn".equals(temp.trim().toLowerCase()))
            level = LogLevel.WARN;
         else if ("error".equals(temp.trim().toLowerCase()))
            level = LogLevel.ERROR;
      }

      if (System.getProperty(ENV_MAX_BACKUP_FILES) != null)
      {
         try 
         {
            String temp = System.getProperty(ENV_MAX_BACKUP_FILES);
            maxBackupFiles = Integer.parseInt(temp.trim());
         } 
         catch (NumberFormatException e)
         {
            e.printStackTrace();
         }
      }

      if (System.getProperty(ENV_MAX_FILE_SIZE) != null)
      {
         try 
         {
            String temp = System.getProperty(ENV_MAX_FILE_SIZE);
            maxFileSize = Integer.parseInt(temp.trim());
         } 
         catch (NumberFormatException e)
         {
            e.printStackTrace();
         }
      }

      if (System.getProperty(ENV_LOG_ROTATION_TYPE) != null)
      {
         String temp = System.getProperty(ENV_LOG_ROTATION_TYPE);
         if ("daily".equals(temp.trim().toLowerCase()))
            rotationType = LogRotationType.DAILY;
         else if ("size".equals(temp.trim().toLowerCase()))
            rotationType = LogRotationType.SIZE;
      }

      if (System.getProperty(ENV_INCLUDE_PACKAGE_NAME) != null)
      {
         String temp = System.getProperty(ENV_INCLUDE_PACKAGE_NAME);
         includePackageName =  Boolean.valueOf(temp.trim()).booleanValue();
      }
   }
   
   /**
    * Output the values used for the run. 
    */
   private static void handleOutputtingValues()
   {
      SimpleLogger.info("Log Path And FileName: " + pathAndFileName);
      SimpleLogger.info("Log Level:             " + level);
      SimpleLogger.info("Log Rotation Type:     " + rotationType);
      SimpleLogger.info("Max. Backup Files:     " + maxBackupFiles);
      SimpleLogger.info("Max. Log File Size:    " + maxFileSize);
      SimpleLogger.info("Include Package Name:  " + includePackageName);
   }
   
   /**
    * Close the log file and reset the JVM's output to standard out and error. 
    */
   public static void close()
   {
      try
      {
         if (ps != null)
            ps.close();
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
      
      try
      {
         if (fos != null)
            fos.close();
      }
      catch (IOException e)
      {
         e.printStackTrace();
      }

      /*
       * Reset the system stdout & errout back to default console. 
       */
      System.setOut(new PrintStream(new FileOutputStream(FileDescriptor.out)));
      System.setErr(new PrintStream(new FileOutputStream(FileDescriptor.err)));
   }
   
   /**
    * Set whether or not to include the package name with the class name.  Default value is true.
    * @param flag Indicate whether the package name should be included with the class name.
    */
   public static void setIncludePackageName(boolean flag)
   {
      includePackageName = flag;
   }
   
   /**
    * Set the log level.  Default value is LogLevel.INFO
    * @param ll the log level to be used.
    */
   public static void setLevel(LogLevel ll)
   {
      level = ll;
   }
   
   /**
    * Set the maximum number of backup files.  Default value is 9.
    * @param maxCount the maximum number of backup files to be kept.
    */
   public static void setMaxBackupFiles(int maxCount)
   {
      maxBackupFiles = maxCount;
   }
   
   /**
    * Set the maximum log file size.  Default value is 10485760 (10MB).
    * @param maxSize the maximum size of the log file before rotation.  Only valid in log level is set to LogRotationType.SIZE
    */
   public static void setMaxLogFileSize(int maxSize)
   {
      maxFileSize = maxSize;
   }
   
   /**
    * Set the log rotation type. Default value is LogRotationType.DAILY
    * @param rt the type of log rotation to be used.
    */
   public static void setRotationType(LogRotationType rt)
   {
      rotationType = rt;
   }
   
   /**
    * Write the data to the log file if the logLevel is at least ERROR.
    * @param data the data to be written to the log file.
    */
   public static void error(Object data)
   {
      if ( (level == LogLevel.ERROR) || (level == LogLevel.WARN) || (level == LogLevel.INFO) || (level == LogLevel.DEBUG) )
         writeDataToFile(data, "ERROR");
   }
   
   /**
    * Write the data to the log file if the logLevel is at least WARN.
    * @param data the data to be written to the log file.
    */
   public static void warn(Object data)
   {
      if ( (level == LogLevel.WARN) || (level == LogLevel.INFO) || (level == LogLevel.DEBUG) )
         writeDataToFile(data, "WARN ");
   }
   
   /**
    * Write the data to the log file if the logLevel is at least INFO.
    * @param data the data to be written to the log file.
    */
   public static void info(Object data)
   {
      if ( (level == LogLevel.INFO) || (level == LogLevel.DEBUG) )
         writeDataToFile(data, "INFO ");
   }
   
   /**
    * Write the data to the log file if the logLevel is at least INFO.
    * @param data the data to be written to the log file.
    */
   public static void debug(Object data)
   {
      if (level == LogLevel.DEBUG)
         writeDataToFile(data, "DEBUG");
   }
   
   /**
    * Log the data to the file.
    * @param data the data to be written to the log file.
    */
   private static synchronized void writeDataToFile(Object data, String label)
   {
      rollLog();

      synchronized (pathAndFileName)
      {
         String className = Thread.currentThread().getStackTrace()[3].getClassName();
   
         if (!includePackageName)
         {
            // Remove the package info.
            if ( (className != null) && (className.lastIndexOf('.') != -1) )
               className = className.substring(className.lastIndexOf('.')+1);
         }
   
         try
         {
            System.out.println(LocalDateTime.now().format(LOGGER_TIMESTAMP) + " " + label +" ("+className+"."+Thread.currentThread().getStackTrace()[3].getMethodName()+") "+data);
         }
         catch(Exception e)
         {
            e.printStackTrace();
         }
      }
   }
   
   /**
    * Convert the data to a HEX table and write it to the log file if the logLevel is at least ERROR.
    * @param title the title for the HEX dump.
    * @param data the data, convert to HEX, to be written to the log file.
    */
   public static void errorDump(String title, byte[] data)
   {
      if ( (level == LogLevel.ERROR) || (level == LogLevel.WARN) || (level == LogLevel.INFO) || (level == LogLevel.DEBUG) )
         writeHexDataToFile(title, data, "ERROR");
   }
   
   /**
    * Convert the data to a HEX table and write it to the log file if the logLevel is at least WARN.
    * @param title the title for the HEX dump.
    * @param data the data, convert to HEX, to be written to the log file.
    */
   public static void warnDump(String title, byte[] data)
   {
      if ( (level == LogLevel.WARN) || (level == LogLevel.INFO) || (level == LogLevel.DEBUG) )
         writeHexDataToFile(title, data, "WARN ");
   }
   
   /**
    * Convert the data to a HEX table and write it to the log file if the logLevel is at least INFO.
    * @param title the title for the HEX dump.
    * @param data the data, convert to HEX, to be written to the log file.
    */
   public static void infoDump(String title, byte[] data)
   {
      if ( (level == LogLevel.INFO) || (level == LogLevel.DEBUG) )
         writeHexDataToFile(title, data, "INFO ");
   }
   
   /**
    * Convert the data to a HEX table and write it to the log file if the logLevel is at least DEBUG.
    * @param title the title for the HEX dump.
    * @param data the data, convert to HEX, to be written to the log file.
    */
   public static void debugDump(String title, byte[] data)
   {
      if (level == LogLevel.DEBUG)
         writeHexDataToFile(title, data, "DEBUG");
   }

   /**
    * This method will dump the byte array in hexadecimal format to the log file.
    * @param title the title for the HEX dump.
    * @param data the data, convert to HEX, to be written to the log file.
    */
   private static synchronized void writeHexDataToFile(String title, byte[] data, String label)
   {
      rollLog();

      synchronized (pathAndFileName)
      {
         int widthSize = 32;
         int endPt = widthSize;
         int len = data.length;
         byte[] tempB = new byte[widthSize];

         String className = Thread.currentThread().getStackTrace()[3].getClassName();
   
         if (!includePackageName)
         {
            // Remove the package info.
            if ( (className != null) && (className.lastIndexOf('.') != -1) )
               className = className.substring(className.lastIndexOf('.')+1);
         }
         
         String recHeader = LocalDateTime.now().format(LOGGER_TIMESTAMP) + " " + label +" ("+className+"."+Thread.currentThread().getStackTrace()[3].getMethodName()+") ";
         try
         {
            System.out.println(recHeader + title + " -->>");
   
            for (int i=0; i < len; i+=widthSize)
            {
               if (i+widthSize >= len)
                  endPt = len - i;
   
               for (int j=0; j < endPt; j++)
                  tempB[j] = data[i+j];
   
               System.out.println(recHeader + formatHex(tempB, (i+widthSize < len?widthSize:len-i), i, widthSize ));
               
            }
            
            System.out.println(recHeader + "<<-----");
         }
         catch (Exception e)
         {
            e.printStackTrace();
         }
      }
   }

   /**
    * Format an array of bytes into a hex display
    * @param src the source byte array
    * @param len the length of the data in the byte array
    * @param index the position of the data within the original byte array
    * @param width the width of the HEX dump of each line
    * @return
    */
   private static String formatHex(byte[] src, int lenSrc, int index, int width)
   {
      int i,j;
      int g = width / 4; /* number of groups of 4 bytes */
      int d = g * 9;     /* hex display width */
      StringBuilder sb = new StringBuilder();

      if ( (src == null) ||
           (lenSrc < 1) || (lenSrc > width) ||
           (index < 0) ||
           (g % 4 != 0) ||   /* only allow width of 16 / 32 / 48 etc. */
           (d < 36) )
      {
         return "";
      }

      String hdr = Integer.toHexString(index);
      if (hdr.length() <= 6)
         sb.append("000000".substring(0, 6 - hdr.length()) + hdr + ": ");
      else
         sb.append(hdr + ": ");

      /* hex display 4 by 4 */
      for(i=0; i < lenSrc; i++)
      {
         sb.append("" + HEX.charAt((src[i] & 0xF0) >> 4) + HEX.charAt((src[i] & 0x0F)));

         if (((i+1) % 4) == 0)
            sb.append(" ");
      }

      /* blank fill hex area if we do not have "width" bytes */
      if (lenSrc < width)
      {
         j = d - (lenSrc*2) - (lenSrc / 4);
         for(i=0; i < j; i++)
            sb.append(" ");
      }

      /* character display */
      sb.append(" ");
      for (i=0; i < lenSrc; i++)
      {
         if(Character.isISOControl((char)src[i]))
            sb.append(".");
         else
            sb.append((char)src[i]);
      }

      /* blank fill character area if we do not have "width" bytes */
      if (lenSrc < width)
      {
         j = width - lenSrc;
         for(i=0; i < j; i++)
            sb.append(" ");
      }

      return sb.toString();
   }

   /**
    * Check if the log file needs to be rolled, if so, roll it.
    */
   private static synchronized void rollLog()
   {
      synchronized (pathAndFileName)
      {
         if ((new File(pathAndFileName)).exists())
         {
            boolean moveFiles = false;
            
            if (rotationType == LogRotationType.DAILY)
            {
               try
               {
                  if (julianDate != Integer.parseInt(getLastModified((new File(pathAndFileName))).format(JULIAN_DATE)))
                     moveFiles = true;
               }
               catch (Exception e)
               {
                  e.printStackTrace();
               }
            }
            else // LogRotationType.SIZE
            {
               try
               {
                  if (fos.getChannel().size() > maxFileSize)
                     moveFiles = true;
               }
               catch (Exception e)
               {
                  e.printStackTrace();
               }
            }
            
            if (moveFiles)
            {
               close();  // close the log file, so that it can be moved.
               moveLogFiles();
               handleSettingPrintStream();
            }
         }
      }
   }

   /**
    * Move the log file(s)
    */
   private static synchronized void moveLogFiles()
   {
      synchronized (pathAndFileName)
      {
         String oldBackupLogFile = pathAndFileName + "." + maxBackupFiles;
         if ((new File(oldBackupLogFile)).exists())
            (new File(oldBackupLogFile)).delete();
   
         for (int i=maxBackupFiles; i >= 2; i--)
         {
            oldBackupLogFile = pathAndFileName + "." + (i-1);
            String newBackupLogFile = pathAndFileName + "." + i;
            if ((new File(oldBackupLogFile)).exists())
               (new File(oldBackupLogFile)).renameTo(new File(newBackupLogFile));
         }
   
         oldBackupLogFile = pathAndFileName;
         String newBackupLogFile = pathAndFileName + ".1";
         if ((new File(oldBackupLogFile)).exists())
            (new File(oldBackupLogFile)).renameTo(new File(newBackupLogFile));
      }
   }

   /**
    * Retrieve the file's last modified time.
    * @param file the file abstract
    * @return ldt the last modified date of the file 
    * @throws IOException
    */
   private static LocalDateTime getLastModified(File file) throws IOException
   {
      Path path = Paths.get(file.getPath());
      BasicFileAttributeView basicfile = Files.getFileAttributeView(path, BasicFileAttributeView.class, LinkOption.NOFOLLOW_LINKS);
      BasicFileAttributes attr = basicfile.readAttributes();
      long date = attr.lastModifiedTime().toMillis();
      Instant instant = Instant.ofEpochMilli(date);
      return LocalDateTime.ofInstant(instant, ZoneId.systemDefault());
   }
}
